
/*
   Defines the BLAS based vector operations. Code shared by parallel
  and sequential vectors.
*/

#include <petsc-private/vecimpl.h> 
#include <../src/vec/vec/impls/dvecimpl.h> 
#include <petscblaslapack.h>
#include <petscthreadcomm.h>

#if defined(PETSC_THREADCOMM_ACTIVE)
PetscErrorCode VecDot_kernel(PetscInt thread_id,Vec xin,Vec yin,PetscThreadCommRedCtx red)
{
  PetscErrorCode       ierr;
  PetscInt             *trstarts=xin->map->trstarts;
  PetscInt             start,end,n;
  PetscBLASInt         one = 1,bn;
  const PetscScalar    *xa,*ya;
  PetscScalar          z_loc;

  ierr = VecGetArrayRead(xin,&xa);CHKERRQ(ierr);
  ierr = VecGetArrayRead(yin,&ya);CHKERRQ(ierr);
  start = trstarts[thread_id];
  end   = trstarts[thread_id+1];
  n     = end-start;
  bn = PetscBLASIntCast(n);
  /* arguments ya, xa are reversed because BLAS complex conjugates the first argument, PETSc the second */
  z_loc = BLASdot_(&bn,ya+start,&one,xa+start,&one);

  ierr = PetscThreadReductionKernelBegin(thread_id,red,(void*)&z_loc);CHKERRQ(ierr);

  ierr = VecRestoreArrayRead(xin,&xa);CHKERRQ(ierr);
  ierr = VecRestoreArrayRead(yin,&ya);CHKERRQ(ierr);
  return 0;
}  

#undef __FUNCT__  
#define __FUNCT__ "VecDot_Seq"
PetscErrorCode VecDot_Seq(Vec xin,Vec yin,PetscScalar *z)
{
  PetscErrorCode        ierr;
  PetscThreadCommRedCtx red;

  PetscFunctionBegin;
  ierr = PetscThreadReductionBegin(((PetscObject)xin)->comm,THREADCOMM_SUM,PETSC_SCALAR,&red);CHKERRQ(ierr);
  ierr = PetscThreadCommRunKernel(((PetscObject)xin)->comm,(PetscThreadKernel)VecDot_kernel,3,xin,yin,red);CHKERRQ(ierr);
  ierr = PetscThreadReductionEnd(red,z);CHKERRQ(ierr);
  if (xin->map->n > 0) {
    ierr = PetscLogFlops(2.0*xin->map->n-1);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
#else
#undef __FUNCT__  
#define __FUNCT__ "VecDot_Seq"
PetscErrorCode VecDot_Seq(Vec xin,Vec yin,PetscScalar *z)
{
  const PetscScalar *ya,*xa;
  PetscBLASInt      one = 1,bn;
  PetscErrorCode    ierr;

  PetscFunctionBegin;
#if !defined(PETSC_HAVE_OPENMP)
  bn = PetscBLASIntCast(xin->map->n);
#endif
  ierr = VecGetArrayRead(xin,&xa);CHKERRQ(ierr);
  ierr = VecGetArrayRead(yin,&ya);CHKERRQ(ierr);
#if defined(PETSC_HAVE_OPENMP)
  {
    PetscScalar sum = 0.0;
    VecOMPParallelBegin(xin, shared(xa,ya,one) reduction(+:sum) default(none));
    {
      PetscBLASInt len, offset;
      VecOMPGetBLASOffsets(xin, &len, &offset);
      sum += BLASdot_(&len, xa + offset, &one, ya + offset, &one);
    }
    VecOMPParallelEnd();
    *z = sum;
  }
#else
  *z = BLASdot_(&bn,xa,&one,ya,&one);
#endif
  ierr = VecRestoreArrayRead(xin,&xa);CHKERRQ(ierr);
  ierr = VecRestoreArrayRead(yin,&ya);CHKERRQ(ierr);
  if (xin->map->n > 0) {
    ierr = PetscLogFlops(2.0*xin->map->n-1);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
#endif

#undef __FUNCT__  
#define __FUNCT__ "VecTDot_Seq"
PetscErrorCode VecTDot_Seq(Vec xin,Vec yin,PetscScalar *z)
{
  const PetscScalar *ya,*xa;
  PetscBLASInt      one = 1,bn;
  PetscErrorCode    ierr;

  PetscFunctionBegin;
#if !defined(PETSC_HAVE_OPENMP)
  bn = PetscBLASIntCast(xin->map->n);
#endif
  ierr = VecGetArrayRead(xin,&xa);CHKERRQ(ierr);
  ierr = VecGetArrayRead(yin,&ya);CHKERRQ(ierr);
#if defined(PETSC_HAVE_OPENMP)
 {
   PetscScalar sum = 0.0;
   VecOMPParallelBegin(xin, shared(xa,ya,one) reduction(+:sum) default(none));
   {
     PetscBLASInt len, offset;
     VecOMPGetBLASOffsets(xin, &len, &offset);
     sum += BLASdotu_(&len, xa + offset, &one, ya + offset, &one);
   }
   VecOMPParallelEnd();
   *z = sum;
 }
#else
  *z = BLASdotu_(&bn,xa,&one,ya,&one);
#endif
  ierr = VecRestoreArrayRead(xin,&xa);CHKERRQ(ierr);
  ierr = VecRestoreArrayRead(yin,&ya);CHKERRQ(ierr);
  if (xin->map->n > 0) {
    ierr = PetscLogFlops(2.0*xin->map->n-1);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "VecScale_Seq"
PetscErrorCode VecScale_Seq(Vec xin, PetscScalar alpha)
{
  PetscErrorCode ierr;
  PetscBLASInt   one = 1;
#if !defined(PETSC_HAVE_OPENMP)
  PetscBLASInt   bn = PetscBLASIntCast(xin->map->n);
#endif

  PetscFunctionBegin;
  if (alpha == (PetscScalar)0.0) {
    ierr = VecSet_Seq(xin,alpha);CHKERRQ(ierr);
  } else if (alpha != (PetscScalar)1.0) {
    PetscScalar a = alpha,*xarray;
    ierr = VecGetArray(xin,&xarray);CHKERRQ(ierr);
#if defined(PETSC_HAVE_OPENMP)
    VecOMPParallelBegin(xin, shared(xarray,a,one) default(none));
    {
      PetscBLASInt len, offset;
      VecOMPGetBLASOffsets(xin, &len, &offset);
      BLASscal_(&len, &a, xarray + offset, &one);
    }
    VecOMPParallelEnd();
#else
    BLASscal_(&bn,&a,xarray,&one);
#endif
    ierr = VecRestoreArray(xin,&xarray);CHKERRQ(ierr);
  }
  ierr = PetscLogFlops(xin->map->n);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#if defined(PETSC_THREADCOMM_ACTIVE)
PetscErrorCode VecAXPY_kernel(PetscInt thread_id,Vec yin,PetscScalar *alpha_p,Vec xin)
{
  PetscErrorCode    ierr;
  const PetscScalar *xarray;
  PetscScalar       *yarray;
  PetscBLASInt      one=1,bn;
  PetscInt          *trstarts=yin->map->trstarts,start,end,n;
  PetscScalar       alpha = *alpha_p;

  start = trstarts[thread_id];
  end   = trstarts[thread_id+1];
  n     = end - start;
  bn = PetscBLASIntCast(n);
  ierr = VecGetArrayRead(xin,&xarray);CHKERRQ(ierr);
  ierr = VecGetArray(yin,&yarray);CHKERRQ(ierr);
  BLASaxpy_(&bn,&alpha,xarray+start,&one,yarray+start,&one);
  ierr = VecRestoreArrayRead(xin,&xarray);CHKERRQ(ierr);
  ierr = VecRestoreArray(yin,&yarray);CHKERRQ(ierr);

  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "VecAXPY_Seq"
PetscErrorCode VecAXPY_Seq(Vec yin,PetscScalar alpha,Vec xin)
{
  PetscErrorCode    ierr;

  PetscFunctionBegin;
  /* assume that the BLAS handles alpha == 1.0 efficiently since we have no fast code for it */
  if (alpha != (PetscScalar)0.0) {
    PetscScalar *scalar;
    ierr = PetscThreadCommGetScalars(((PetscObject)yin)->comm,&scalar,PETSC_NULL,PETSC_NULL);CHKERRQ(ierr);
    *scalar = alpha;
    ierr = PetscThreadCommRunKernel(((PetscObject)yin)->comm,(PetscThreadKernel)VecAXPY_kernel,3,yin,scalar,xin);CHKERRQ(ierr);
    ierr = PetscLogFlops(2.0*yin->map->n);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
#else
#undef __FUNCT__  
#define __FUNCT__ "VecAXPY_Seq"
PetscErrorCode VecAXPY_Seq(Vec yin,PetscScalar alpha,Vec xin)
{
  PetscErrorCode    ierr;
  const PetscScalar *xarray;
  PetscScalar       *yarray;
  PetscBLASInt      one = 1;
#if !defined(PETSC_HAVE_OPENMP)
  PetscBLASInt      bn = PetscBLASIntCast(xin->map->n);
#endif

  PetscFunctionBegin;
  /* assume that the BLAS handles alpha == 1.0 efficiently since we have no fast code for it */
  if (alpha != (PetscScalar)0.0) {
    ierr = VecGetArrayRead(xin,&xarray);CHKERRQ(ierr);
    ierr = VecGetArray(yin,&yarray);CHKERRQ(ierr);
#if defined(PETSC_HAVE_OPENMP)
    VecOMPParallelBegin(yin, shared(alpha, xarray, one, yarray) default(none));
    {
      PetscBLASInt len, offset;
      VecOMPGetBLASOffsets(yin, &len, &offset);
      BLASaxpy_(&len, &alpha, xarray + offset, &one, yarray + offset, &one);
    }
    VecOMPParallelEnd();
#else
    BLASaxpy_(&bn,&alpha,xarray,&one,yarray,&one);
#endif
    ierr = VecRestoreArrayRead(xin,&xarray);CHKERRQ(ierr);
    ierr = VecRestoreArray(yin,&yarray);CHKERRQ(ierr);
    ierr = PetscLogFlops(2.0*yin->map->n);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}
#endif

#undef __FUNCT__  
#define __FUNCT__ "VecAXPBY_Seq"
PetscErrorCode VecAXPBY_Seq(Vec yin,PetscScalar alpha,PetscScalar beta,Vec xin)
{
  PetscErrorCode    ierr;
  PetscInt          i;
  const PetscScalar *xx;
  PetscScalar       *yy,a = alpha,b = beta;

  PetscFunctionBegin;
  if (a == (PetscScalar)0.0) {
    ierr = VecScale_Seq(yin,beta);CHKERRQ(ierr);
  } else if (b == (PetscScalar)1.0) {
    ierr = VecAXPY_Seq(yin,alpha,xin);CHKERRQ(ierr);
  } else if (a == (PetscScalar)1.0) {
    ierr = VecAYPX_Seq(yin,beta,xin);CHKERRQ(ierr);
  } else if (b == (PetscScalar)0.0) {
    ierr = VecGetArrayRead(xin,&xx);CHKERRQ(ierr);
    ierr = VecGetArray(yin,(PetscScalar**)&yy);CHKERRQ(ierr);
    VecOMPParallelBegin(yin, private(i) shared(xx, yy, a) default(none));
    for (i=__start; i<__end; i++) {
      yy[i] = a*xx[i];
    }
    VecOMPParallelEnd();
    ierr = VecRestoreArrayRead(xin,&xx);CHKERRQ(ierr);
    ierr = VecRestoreArray(yin,(PetscScalar**)&yy);CHKERRQ(ierr);
    ierr = PetscLogFlops(xin->map->n);CHKERRQ(ierr);
  } else {
    ierr = VecGetArrayRead(xin,&xx);CHKERRQ(ierr);
    ierr = VecGetArray(yin,(PetscScalar**)&yy);CHKERRQ(ierr);
    VecOMPParallelBegin(yin, private(i) shared(xx, yy, a, b) default(none));
    for (i=__start; i<__end; i++) {
      yy[i] = a*xx[i] + b*yy[i];
    }
    VecOMPParallelEnd();
    ierr = VecRestoreArrayRead(xin,&xx);CHKERRQ(ierr);
    ierr = VecRestoreArray(yin,(PetscScalar**)&yy);CHKERRQ(ierr);
    ierr = PetscLogFlops(3.0*xin->map->n);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "VecAXPBYPCZ_Seq"
PetscErrorCode VecAXPBYPCZ_Seq(Vec zin,PetscScalar alpha,PetscScalar beta,PetscScalar gamma,Vec xin,Vec yin)
{
  PetscErrorCode     ierr;
  PetscInt           n = zin->map->n,i;
  const PetscScalar  *yy,*xx;
  PetscScalar        *zz;

  PetscFunctionBegin;
  ierr = VecGetArrayRead(xin,&xx);CHKERRQ(ierr);
  ierr = VecGetArrayRead(yin,&yy);CHKERRQ(ierr);
  ierr = VecGetArray(zin,&zz);CHKERRQ(ierr);
  if (alpha == (PetscScalar)1.0) {
    VecOMPParallelBegin(zin, private(i) shared(xx, yy, zz, beta, gamma) default(none));
    for (i=__start; i<__end; i++) {
      zz[i] = xx[i] + beta*yy[i] + gamma*zz[i];
    }
    VecOMPParallelEnd();
    ierr = PetscLogFlops(4.0*n);CHKERRQ(ierr);
  } else if (gamma == (PetscScalar)1.0) {
    VecOMPParallelBegin(zin, private(i) shared(xx, yy, zz, alpha, beta) default(none));
    for (i=__start; i<__end; i++) {
      zz[i] = alpha*xx[i] + beta*yy[i] + zz[i];
    }
    VecOMPParallelEnd();
    ierr = PetscLogFlops(4.0*n);CHKERRQ(ierr);
  } else if (gamma == (PetscScalar)0.0) {
    VecOMPParallelBegin(zin, private(i) shared(xx, yy, zz, alpha, beta) default(none));
    for (i=__start; i<__end; i++) {
      zz[i] = alpha*xx[i] + beta*yy[i];
    }
    VecOMPParallelEnd();
    ierr = PetscLogFlops(3.0*n);CHKERRQ(ierr);
  } else {
    VecOMPParallelBegin(zin, private(i) shared(xx, yy, zz, alpha, beta, gamma) default(none));
    for (i=__start; i<__end; i++) {
      zz[i] = alpha*xx[i] + beta*yy[i] + gamma*zz[i];
    }
    VecOMPParallelEnd();
    ierr = PetscLogFlops(5.0*n);CHKERRQ(ierr);
  }
  ierr = VecRestoreArrayRead(xin,&xx);CHKERRQ(ierr);
  ierr = VecRestoreArrayRead(yin,&yy);CHKERRQ(ierr);
  ierr = VecRestoreArray(zin,&zz);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
